import os
import sys

from django.core.management import execute_from_command_line


if __name__ == '__main__':
    os.environ['DJANGO_SETTINGS_MODULE'] = 'settings'
    execute_from_command_line(sys.argv)



from django.conf.urls import url

import app.views as views


urlpatterns = [
    url(r'^$', views.home, name='home'),
    url(r'^user', views.user, name='user'),
    url(r'^error', views.error, name='error'),
    url(r'^bad', views.bad_request, name='bad'),
    url(r'^nested', views.nested, name='nested'),
]



ROOT_URLCONF = 'urls'

SECRET_KEY = 'euowhnfckwnqopifnkadc;wmfjgwkns'

INSTALLED_APPS = (
    'app',
)

MIDDLEWARE_CLASSES = (
    'opentracing_utils.OpenTracingHttpMiddleware',
)

MIDDLEWARE = MIDDLEWARE_CLASSES




import opentracing

import pytest

import six

from mock import MagicMock

from opentracing.ext import tags
from basictracer import BasicTracer

from ..conftest import Recorder

from opentracing_utils import extract_span_from_django_request


@pytest.mark.skipif(six.PY2, reason='')
def skip_span(request, view_func, view_args, view_kwargs):
    if view_func.__name__ == 'home':
        return True
    return False


def operation_name(request, view_func, view_args, view_kwargs):
    return 'custom-op-name'


def get_recorder():
    recorder = Recorder()
    t = BasicTracer(recorder=recorder)
    t.register_required_propagators()
    opentracing.tracer = t

    return recorder


@pytest.mark.skipif(six.PY2, reason='')
@pytest.mark.parametrize('url,method,status', (
    ('/', 'GET', 200),
    ('/', 'POST', 200),
    ('/notfound', 'GET', 404),
))
def test_request_home(client, url, method, status):
    recorder = get_recorder()

    m = client.get if method == 'GET' else client.post
    response = m(url)
    if status != 404:
        assert response.content == b'TRACED'

        assert len(recorder.spans) == 1

        assert recorder.spans[0].operation_name == 'home'
        assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
        assert recorder.spans[0].tags[tags.HTTP_URL] == url
        assert recorder.spans[0].tags[tags.HTTP_METHOD] == method
        assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == status


@pytest.mark.skipif(six.PY2, reason='')
def test_request_nested(client):
    recorder = get_recorder()

    response = client.get('/nested')
    assert response.content == b'NESTED'

    assert len(recorder.spans) == 2

    assert recorder.spans[1].operation_name == 'nested'
    assert recorder.spans[1].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[1].tags[tags.HTTP_URL] == '/nested'
    assert recorder.spans[1].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[1].tags[tags.HTTP_STATUS_CODE] == 200

    assert recorder.spans[0].operation_name == 'nested_call'
    assert recorder.spans[0].tags['nested'] is True
    assert recorder.spans[0].context.trace_id == recorder.spans[1].context.trace_id
    assert recorder.spans[0].parent_id == recorder.spans[1].context.span_id


@pytest.mark.skipif(six.PY2, reason='')
def test_request_user(client):
    recorder = get_recorder()

    response = client.get('/user')
    assert response.content == b'USER'

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'user'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/user'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 200


@pytest.mark.skipif(six.PY2, reason='')
def test_request_error(client):
    recorder = get_recorder()

    with pytest.raises(RuntimeError):
        client.get('/error')

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'error'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/error'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 500
    assert recorder.spans[0].tags['error'] is True


@pytest.mark.skipif(six.PY2, reason='')
@pytest.mark.parametrize('op_name', ('tests.test_django.test_django_middleware.skip_span', skip_span))
def test_request_skip_span(client, settings, op_name):
    recorder = get_recorder()

    settings.OPENTRACING_UTILS_SKIP_SPAN_CALLABLE = op_name

    response = client.get('/')
    assert response.content == b'TRACED'

    assert len(recorder.spans) == 0

    response = client.get('/user')
    assert response.content == b'USER'

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'user'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/user'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 200


@pytest.mark.skipif(six.PY2, reason='')
@pytest.mark.parametrize('op_name', ('tests.test_django.test_django_middleware.operation_name', operation_name))
def test_request_operation_name(client, settings, op_name):
    recorder = get_recorder()

    settings.OPENTRACING_UTILS_OPERATION_NAME_CALLABLE = op_name

    response = client.get('/user')
    assert response.content == b'USER'

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'custom-op-name'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/user'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 200


@pytest.mark.skipif(six.PY2, reason='')
@pytest.mark.parametrize('default_tags', (
    {},
    '',
    None,
    [],
    {'tag-1': 'val-1', 'tag-2': 'val-2'},
))
def test_request_default_tags(client, settings, default_tags):
    recorder = get_recorder()

    settings.OPENTRACING_UTILS_DEFAULT_TAGS = default_tags

    response = client.get('/user')
    assert response.content == b'USER'

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'user'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/user'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 200

    if type(default_tags) is dict:
        for t, v in default_tags.items():
            recorder.spans[0].tags[t] == v


@pytest.mark.skipif(six.PY2, reason='')
@pytest.mark.parametrize('error_4xx', (True, False))
def test_request_ignore_4xx(client, settings, error_4xx):
    recorder = get_recorder()

    settings.OPENTRACING_UTILS_ERROR_4XX = error_4xx

    client.get('/bad')

    assert len(recorder.spans) == 1

    assert recorder.spans[0].operation_name == 'bad_request'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/bad'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 400

    assert ('error' in recorder.spans[0].tags) is error_4xx


@pytest.mark.skipif(six.PY2, reason='')
def test_request_propagated(monkeypatch, client):
    recorder = get_recorder()

    propagated_span = MagicMock()
    propagated_span.context.trace_id = 123
    propagated_span.context.span_id = 123456

    def extract(fmt, carrier):
        assert 'x-trace-id' in carrier
        return propagated_span

    monkeypatch.setattr('opentracing.tracer.extract', extract)

    response = client.get('/', HTTP_X_TRACE_ID=123)  # HTTP_X_TRACE_ID is only used to test headers transformation!
    assert response.content == b'TRACED'

    assert recorder.spans[0].context.trace_id == propagated_span.context.trace_id
    assert recorder.spans[0].parent_id == propagated_span.context.span_id

    assert recorder.spans[0].operation_name == 'home'
    assert recorder.spans[0].tags[tags.COMPONENT] == 'django'
    assert recorder.spans[0].tags[tags.HTTP_URL] == '/'
    assert recorder.spans[0].tags[tags.HTTP_METHOD] == 'GET'
    assert recorder.spans[0].tags[tags.HTTP_STATUS_CODE] == 200


@pytest.mark.skipif(six.PY2, reason='')
def test_extract_span():
    request = MagicMock()

    request.current_span = '1'

    assert '1' == extract_span_from_django_request(request)



from django.http import HttpResponse

from opentracing_utils import trace, extract_span_from_django_request, extract_span_from_kwargs


def home(request):
    return HttpResponse('TRACED')


def user(request):
    return HttpResponse('USER')


def error(request):
    raise RuntimeError('Failed request')


def bad_request(request):
    return HttpResponse(status=400)


@trace(span_extractor=extract_span_from_django_request, operation_name='nested_call', pass_span=True)
def nested(request, *args, **kwargs):
    current_span = extract_span_from_kwargs(**kwargs)
    current_span.set_tag('nested', True)

    return HttpResponse('NESTED')



"""
Django OpenTracing middleware.

This is a slightly modified version of the middleware implemented in:
https://github.com/opentracing-contrib/python-django [BSD 3-Clause]
"""
import traceback
import opentracing


settings = None  # noqa

try:
    from django.conf import settings
    from django.utils.deprecation import MiddlewareMixin
except ImportError:  # pragma: no cover
    MiddlewareMixin = object

try:
    from django.utils.module_loading import import_string
except ImportError:  # pragma: no cover
    try:
        from django.utils.module_loading import import_by_path as import_string
    except ImportError:  # pragma: no cover
        pass

from opentracing.ext import tags as ot_tags

from opentracing_utils.common import sanitize_url


class OpenTracingHttpMiddleware(MiddlewareMixin):

    def __init__(self, get_response=None):
        self.get_response = get_response

        self._default_tags = getattr(settings, 'OPENTRACING_UTILS_DEFAULT_TAGS', {})
        if type(self._default_tags) is not dict:
            self._default_tags = {}

        error_4xx = bool(getattr(settings, 'OPENTRACING_UTILS_ERROR_4XX', True))
        self._min_error_code = 400 if error_4xx else 500

        op_name_str = getattr(settings, 'OPENTRACING_UTILS_OPERATION_NAME_CALLABLE', '')
        if callable(op_name_str):
            self._op_name_callable = op_name_str
        else:
            self._op_name_callable = import_string(op_name_str) if op_name_str else None

        skip_span_str = getattr(settings, 'OPENTRACING_UTILS_SKIP_SPAN_CALLABLE', '')
        if callable(skip_span_str):
            self._skip_span_callable = skip_span_str
        else:
            self._skip_span_callable = import_string(skip_span_str) if skip_span_str else None

    def process_view(self, request, view_func, view_args, view_kwargs):
        if self._skip_span_callable and self._skip_span_callable(request, view_func, view_args, view_kwargs):
            return

        headers_carrier = self._get_headers(request)

        op_name = (self._op_name_callable(request, view_func, view_args, view_kwargs) if self._op_name_callable
                   else view_func.__name__)

        span = None
        try:
            span_ctx = opentracing.tracer.extract(opentracing.Format.HTTP_HEADERS, headers_carrier)
            span = opentracing.tracer.start_span(operation_name=op_name, child_of=span_ctx)
        except (opentracing.InvalidCarrierException, opentracing.SpanContextCorruptedException):
            span = opentracing.tracer.start_span(operation_name=op_name, tags={'django-no-propagation': True})

        span.set_tag(ot_tags.COMPONENT, 'django')
        span.set_tag(ot_tags.SPAN_KIND, ot_tags.SPAN_KIND_RPC_SERVER)
        span.set_tag(ot_tags.HTTP_METHOD, request.method)
        span.set_tag(ot_tags.HTTP_URL, sanitize_url(request.get_full_path()))

        # Default tags
        for k, v in self._default_tags.items():
            try:
                span.set_tag(k, v)
            except Exception:  # pragma: no cover
                pass

        request.current_span = span

    def process_exception(self, request, exception):
        self._finish_tracing(request, exception=exception)

    def process_response(self, request, response):
        self._finish_tracing(request, response=response)
        return response

    def _get_headers(self, request):
        headers = {}

        for k, v in request.META.items():
            k = k.lower().replace('_', '-')
            k = k.replace('http-', '') if k.startswith('http-') else k  # remove Django magic prefix!
            headers[k] = v

        return headers

    def _finish_tracing(self, request, response=None, exception=None):
        current_span = getattr(request, 'current_span', None)
        if not current_span:
            return

        if response:
            current_span.set_tag(ot_tags.HTTP_STATUS_CODE, response.status_code)

            if response.status_code >= self._min_error_code:
                current_span.set_tag('error', True)

            current_span.finish()
        elif exception:
            current_span.set_tag('error', True)
            current_span.log_kv({
                'error.kind': str(exception),
                'stack': traceback.format_exc(),
            })


def extract_span_from_django_request(request, *args, **kwargs):
    """
    Safe utility function to extract the ``current_span`` from ``HttpRequest``. Compatible with ``@trace`` decorator.
    """
    try:
        return getattr(request, 'current_span', None)
    except Exception:  # pragma: no cover
        pass

    return None  # pragma: no cover
